@@ -56,10 +56,9 @@ static int *VCOPY(const int *valuators, int nvals)
 
 static int applyPressureCurve(WacomDevicePtr pDev, const WacomDeviceStatePtr pState);
 static void commonDispatchDevice(InputInfoPtr pInfo,
-				 const WacomChannelPtr pChannel,
-				 enum WacomSuppressMode suppress);
-static void sendAButton(InputInfoPtr pInfo, int button, int mask,
-			int first_val, int num_vals, int *valuators);
+				 const WacomChannelPtr pChannel);
+static void sendAButton(InputInfoPtr pInfo,  const WacomDeviceState* ds, int button,
+			int mask, int first_val, int num_vals, int *valuators);
 
 /*****************************************************************************
  * Utility functions
@@ -89,16 +88,60 @@ void set_absolute(InputInfoPtr pInfo, Bool absolute)
 		priv->flags &= ~ABSOLUTE_FLAG;
 }
 
+static int wcmButtonPerNotch(WacomDevicePtr priv, int value, int threshold, int btn_positive, int btn_negative)
+{
+	int mode = is_absolute(priv->pInfo);
+	int notches = value / threshold;
+	int button = (notches > 0) ? btn_positive : btn_negative;
+	int i;
+
+	for (i = 0; i < abs(notches); i++) {
+		xf86PostButtonEventP(priv->pInfo->dev, mode, button, 1, 0, 0, 0);
+		xf86PostButtonEventP(priv->pInfo->dev, mode, button, 0, 0, 0, 0);
+	}
+
+	return value % threshold;
+}
+
+static void wcmPanscroll(WacomDevicePtr priv, const WacomDeviceState *ds)
+{
+	WacomCommonPtr common = priv->common;
+	int threshold = common->wcmPanscrollThreshold;
+	int *accumulated_x, *accumulated_y;
+
+	if (!(priv->flags & SCROLLMODE_FLAG) || !(ds->buttons & 1))
+		return;
+
+	/* Tip has gone down down; store state for dragging */
+	if (!(priv->oldState.buttons & 1)) {
+		priv->wcmPanscrollState = *ds;
+		priv->wcmPanscrollState.x = 0;
+		priv->wcmPanscrollState.y = 0;
+		return;
+	}
+
+	accumulated_x = &priv->wcmPanscrollState.x;
+	accumulated_y = &priv->wcmPanscrollState.y;
+
+	*accumulated_x += (ds->x - priv->oldState.x);
+	*accumulated_y += (ds->y - priv->oldState.y);
+
+	DBG(6, priv, "pan x = %d, pan y = %d\n", *accumulated_x, *accumulated_y);
+
+	*accumulated_x = wcmButtonPerNotch(priv, *accumulated_x, threshold, 6, 7);
+	*accumulated_y = wcmButtonPerNotch(priv, *accumulated_y, threshold, 4, 5);
+}
+
 /*****************************************************************************
  * wcmSendButtons --
  *   Send button events by comparing the current button mask with the
  *   previous one.
  ****************************************************************************/
 
-static void wcmSendButtons(InputInfoPtr pInfo, int buttons,
+static void wcmSendButtons(InputInfoPtr pInfo, const WacomDeviceState* ds, int buttons,
 			   int first_val, int num_vals, int *valuators)
 {
-	int button, mask, first_button;
+	unsigned int button, mask, first_button;
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	WacomCommonPtr common = priv->common;
 	DBG(6, priv, "buttons=%d\n", buttons);
@@ -134,9 +177,9 @@ static void wcmSendButtons(InputInfoPtr pInfo, int buttons,
 
 	for (button = first_button; button < WCM_MAX_BUTTONS; button++)
 	{
-		mask = 1 << button;
+		mask = 1u << button;
 		if ((mask & priv->oldState.buttons) != (mask & buttons))
-			sendAButton(pInfo, button, (mask & buttons),
+			sendAButton(pInfo, ds, button, (mask & buttons),
 					first_val, num_vals, valuators);
 	}
 
@@ -166,10 +209,11 @@ static int countPresses(int keybtn, unsigned int* keys, int size)
 	return count;
 }
 
-static void sendAction(InputInfoPtr pInfo, int press,
-		       unsigned int *keys, int nkeys,
+static void sendAction(InputInfoPtr pInfo,  const WacomDeviceState* ds,
+		       int press, unsigned int *keys, int nkeys,
 		       int first_val, int num_val, int *valuators)
 {
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	int i;
 
 	/* Actions only trigger on press, not release */
@@ -186,10 +230,15 @@ static void sendAction(InputInfoPtr pInfo, int press,
 				{
 					int btn_no = (action & AC_CODE);
 					int is_press = (action & AC_KEYBTNPRESS);
-					xf86PostButtonEventP(pInfo->dev,
-							    is_absolute(pInfo), btn_no,
-							    is_press, first_val, num_val,
-							    VCOPY(valuators, num_val));
+					if (btn_no == 1 && (priv->flags & SCROLLMODE_FLAG)) {
+						/* Don't send clicks in scroll mode */
+					}
+					else {
+						xf86PostButtonEventP(pInfo->dev,
+								    is_absolute(pInfo), btn_no,
+								    is_press, first_val, num_val,
+								    VCOPY(valuators, num_val));
+					}
 				}
 				break;
 			case AC_KEY:
@@ -204,6 +253,12 @@ static void sendAction(InputInfoPtr pInfo, int press,
 					wcmDevSwitchModeCall(pInfo,
 							(is_absolute(pInfo)) ? Relative : Absolute); /* not a typo! */
 				break;
+			case AC_PANSCROLL:
+				priv->flags |= SCROLLMODE_FLAG;
+				priv->wcmPanscrollState = *ds;
+				priv->wcmPanscrollState.x = 0;
+				priv->wcmPanscrollState.y = 0;
+				break;
 		}
 	}
 
@@ -240,8 +295,11 @@ static void sendAction(InputInfoPtr pInfo, int press,
 					if (countPresses(key_code, &keys[i], nkeys - i))
 						wcmEmitKeycode(pInfo->dev, key_code, 0);
 				}
+				break;
+			case AC_PANSCROLL:
+				priv->flags &= ~SCROLLMODE_FLAG;
+				break;
 		}
-
 	}
 }
 
@@ -249,8 +307,8 @@ static void sendAction(InputInfoPtr pInfo, int press,
  * sendAButton --
  *   Send one button event, called by wcmSendButtons
  ****************************************************************************/
-static void sendAButton(InputInfoPtr pInfo, int button, int mask,
-			int first_val, int num_val, int *valuators)
+static void sendAButton(InputInfoPtr pInfo, const WacomDeviceState* ds, int button,
+			int mask, int first_val, int num_val, int *valuators)
 {
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 #ifdef DEBUG
@@ -263,7 +321,7 @@ static void sendAButton(InputInfoPtr pInfo, int button, int mask,
 	if (!priv->keys[button][0])
 		return;
 
-	sendAction(pInfo, (mask != 0), priv->keys[button],
+	sendAction(pInfo, ds, (mask != 0), priv->keys[button],
 		   ARRAY_SIZE(priv->keys[button]),
 		   first_val, num_val, valuators);
 }
@@ -345,11 +403,11 @@ TEST_NON_STATIC int getWheelButton(int delta, int action_up, int action_dn)
  * @param num_vals
  * @param valuators
  */
-static void sendWheelStripEvent(unsigned int *action, int nactions, InputInfoPtr pInfo,
-                                int first_val, int num_vals, int *valuators)
+static void sendWheelStripEvent(unsigned int *action, const WacomDeviceState* ds, int nactions,
+				InputInfoPtr pInfo, int first_val, int num_vals, int *valuators)
 {
-	sendAction(pInfo, 1, action, nactions, first_val, num_vals, valuators);
-	sendAction(pInfo, 0, action, nactions, first_val, num_vals, valuators);
+	sendAction(pInfo, ds, 1, action, nactions, first_val, num_vals, valuators);
+	sendAction(pInfo, ds, 0, action, nactions, first_val, num_vals, valuators);
 }
 
 /*****************************************************************************
@@ -361,6 +419,7 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 				 int first_val, int num_vals, int *valuators)
 {
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	WacomCommonPtr common = priv->common;
 	int delta = 0, idx = 0;
 
 	DBG(10, priv, "\n");
@@ -371,7 +430,7 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	if (idx >= 0 && IsPad(priv) && priv->oldState.proximity == ds->proximity)
 	{
 		DBG(10, priv, "Left touch strip scroll delta = %d\n", delta);
-		sendWheelStripEvent(priv->strip_keys[idx], ARRAY_SIZE(priv->strip_keys[idx]),
+		sendWheelStripEvent(priv->strip_keys[idx], ds, ARRAY_SIZE(priv->strip_keys[idx]),
 		                    pInfo, first_val, num_vals, valuators);
 	}
 
@@ -381,7 +440,7 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	if (idx >= 0 && IsPad(priv) && priv->oldState.proximity == ds->proximity)
 	{
 		DBG(10, priv, "Right touch strip scroll delta = %d\n", delta);
-		sendWheelStripEvent(priv->strip_keys[idx], ARRAY_SIZE(priv->strip_keys[idx]),
+		sendWheelStripEvent(priv->strip_keys[idx], ds, ARRAY_SIZE(priv->strip_keys[idx]),
 		                    pInfo, first_val, num_vals, valuators);
 	}
 
@@ -391,27 +450,27 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	if (idx >= 0 && (IsCursor(priv) || IsPad(priv)) && priv->oldState.proximity == ds->proximity)
 	{
 		DBG(10, priv, "Relative wheel scroll delta = %d\n", delta);
-		sendWheelStripEvent(priv->wheel_keys[idx], ARRAY_SIZE(priv->wheel_keys[idx]),
+		sendWheelStripEvent(priv->wheel_keys[idx], ds, ARRAY_SIZE(priv->wheel_keys[idx]),
 		                    pInfo, first_val, num_vals, valuators);
 	}
 
 	/* emulate events for left touch ring */
-	delta = getScrollDelta(ds->abswheel, priv->oldState.abswheel, MAX_PAD_RING, AXIS_INVERT);
+	delta = getScrollDelta(ds->abswheel, priv->oldState.abswheel, common->wcmMaxRing, AXIS_INVERT);
 	idx = getWheelButton(delta, WHEEL_ABS_UP, WHEEL_ABS_DN);
 	if (idx >= 0 && IsPad(priv) && priv->oldState.proximity == ds->proximity)
 	{
 		DBG(10, priv, "Left touch wheel scroll delta = %d\n", delta);
-		sendWheelStripEvent(priv->wheel_keys[idx], ARRAY_SIZE(priv->wheel_keys[idx]),
+		sendWheelStripEvent(priv->wheel_keys[idx], ds, ARRAY_SIZE(priv->wheel_keys[idx]),
 		                    pInfo, first_val, num_vals, valuators);
 	}
 
 	/* emulate events for right touch ring */
-	delta = getScrollDelta(ds->abswheel2, priv->oldState.abswheel2, MAX_PAD_RING, AXIS_INVERT);
+	delta = getScrollDelta(ds->abswheel2, priv->oldState.abswheel2, common->wcmMaxRing, AXIS_INVERT);
 	idx = getWheelButton(delta, WHEEL2_ABS_UP, WHEEL2_ABS_DN);
 	if (idx >= 0 && IsPad(priv) && priv->oldState.proximity == ds->proximity)
 	{
 		DBG(10, priv, "Right touch wheel scroll delta = %d\n", delta);
-		sendWheelStripEvent(priv->wheel_keys[idx], ARRAY_SIZE(priv->wheel_keys[idx]),
+		sendWheelStripEvent(priv->wheel_keys[idx], ds, ARRAY_SIZE(priv->wheel_keys[idx]),
 		                    pInfo, first_val, num_vals, valuators);
 	}
 }
@@ -427,9 +486,12 @@ static void sendCommonEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	int buttons = ds->buttons;
 
+	/* send scrolling events if necessary */
+	wcmPanscroll(priv, ds);
+
 	/* send button events when state changed or first time in prox and button unpresses */
 	if (priv->oldState.buttons != buttons || (!priv->oldState.proximity && !buttons))
-		wcmSendButtons(pInfo,buttons, first_val, num_vals, valuators);
+		wcmSendButtons(pInfo, ds, buttons, first_val, num_vals, valuators);
 
 	/* emulate wheel/strip events when defined */
 	if ( ds->relwheel || (ds->abswheel != priv->oldState.abswheel) || (ds->abswheel2 != priv->oldState.abswheel2) ||
@@ -525,7 +587,7 @@ wcmSendPadEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	else
 	{
 		if (priv->oldState.buttons)
-			wcmSendButtons(pInfo, ds->buttons, first_val, num_vals, valuators);
+			wcmSendButtons(pInfo, ds, ds->buttons, first_val, num_vals, valuators);
 	}
 
 	if (priv->oldState.proximity && !ds->proximity)
@@ -567,7 +629,8 @@ wcmSendNonPadEvents(InputInfoPtr pInfo, const WacomDeviceState *ds,
 						VCOPY(valuators, num_vals));
 
 		/* Move the cursor to where it should be before sending button events */
-		if(!(priv->flags & BUTTONS_ONLY_FLAG))
+		if(!(priv->flags & BUTTONS_ONLY_FLAG) &&
+		   !(priv->flags & SCROLLMODE_FLAG && priv->oldState.buttons & 1))
 		{
 			xf86PostMotionEventP(pInfo->dev, is_absolute(pInfo),
 					     first_val, num_vals,
@@ -593,7 +656,7 @@ wcmSendNonPadEvents(InputInfoPtr pInfo, const WacomDeviceState *ds,
 		/* reports button up when the device has been
 		 * down and becomes out of proximity */
 		if (priv->oldState.buttons)
-			wcmSendButtons(pInfo, buttons, first_val, num_vals, valuators);
+			wcmSendButtons(pInfo, ds, buttons, first_val, num_vals, valuators);
 
 		if (priv->oldState.proximity)
 			xf86PostProximityEventP(pInfo->dev, 0, first_val, num_vals,
@@ -649,6 +712,10 @@ void wcmSendEvents(InputInfoPtr pInfo, const WacomDeviceState* ds)
 		ty = ds->stripy;
 	}
 
+	/* cancel panscroll */
+	if (!ds->proximity)
+		priv->flags &= ~SCROLLMODE_FLAG;
+
 	DBG(7, priv, "[%s] o_prox=%s x=%d y=%d z=%d "
 		"b=%s b=%d tx=%d ty=%d wl=%d wl2=%d rot=%d th=%d\n",
 		pInfo->type_name,
@@ -706,7 +773,8 @@ void wcmSendEvents(InputInfoPtr pInfo, const WacomDeviceState* ds)
 	valuators[3] = v3;
 	valuators[4] = v4;
 	valuators[5] = v5;
-	valuators[6] = v6;
+	if (priv->naxes > 6)
+		valuators[6] = v6;
 
 	if (type == PAD_ID)
 		wcmSendPadEvents(pInfo, ds, 3, priv->naxes - 3, &valuators[3]); /* pad doesn't post x/y/z */
@@ -852,29 +920,38 @@ static Bool check_arbitrated_control(InputInfoPtr pInfo, WacomDeviceStatePtr ds)
 
 	if (IsPad(priv)) {
 		/* Pad may never be the "active" pointer controller */
+		DBG(6, priv, "Event from pad; not yielding pointer control\n.");
 		return FALSE;
 	}
 
 	if (active == NULL || active->oldState.device_id == ds->device_id) {
-		DBG(11, priv, "Same device ID as active; allowing access.\n");
+		DBG(11, priv, "Event from active device; maintaining pointer control.\n");
 		return TRUE;
 	}
-	else if (IsCursor(active) && IsTouch(priv)) {
-		/* Cursor devices are often left idle in range, so allow touch to
-		 * grab control if the tool has not been used for some time.
+	else if (IsCursor(active)) {
+		/* Cursor devices are often left idle in range, so allow other devices
+		 * to grab control if the tool has not been used for some time.
 		 */
-		return (ds->time - active->oldState.time > 100);
+		Bool yield = (ds->time - active->oldState.time > 100) && (active->oldState.buttons == 0);
+		DBG(6, priv, "Currently-active cursor %s idle; %s pointer control.\n",
+		    yield ? "is" : "is not", yield ? "yielding" : "not yielding");
+		return yield;
 	}
-	else if (IsTouch(active) && IsCursor(priv)) {
+	else if (IsCursor(priv)) {
 		/* An otherwise idle cursor may still occasionally jitter and send
-		 * events while the user is making active touches. Do not allow
-		 * the cursor to grab control in this particular case.
+		 * events while the user is actively using other tools or touching
+		 * the device. Do not allow the cursor to grab control in this
+		 * particular case.
 		 */
+		DBG(6, priv, "Event from non-active cursor; not yielding pointer control.\n");
 		return FALSE;
 	}
 	else {
 		/* Non-touch input has priority over touch in general */
-		return !IsTouch(priv);
+		Bool yield = !IsTouch(priv);
+		DBG(6, priv, "Event from non-active %s device; %s pointer control.\n",
+		    yield ? "non-touch" : "touch", yield ? "yielding" : "not yielding");
+		return yield;
 	}
 }
 
@@ -886,15 +963,12 @@ static Bool check_arbitrated_control(InputInfoPtr pInfo, WacomDeviceStatePtr ds)
 void wcmEvent(WacomCommonPtr common, unsigned int channel,
 	const WacomDeviceState* pState)
 {
-	WacomDeviceState* pLast;
 	WacomDeviceState ds;
 	WacomChannelPtr pChannel;
-	enum WacomSuppressMode suppress;
 	InputInfoPtr pInfo;
 	WacomToolPtr tool;
 	WacomDevicePtr priv;
 	pChannel = common->wcmChannel + channel;
-	pLast = &pChannel->valid.state;
 
 	DBG(10, common, "channel = %d\n", channel);
 
@@ -951,21 +1025,6 @@ void wcmEvent(WacomCommonPtr common, unsigned int channel,
 		ds.tilty = 0;
 	}
 
-	/* Optionally filter values only while in proximity */
-	if (ds.proximity && ds.device_type != PAD_ID)
-	{
-		/* Start filter fresh when entering proximity */
-		if (!pLast->proximity)
-			wcmResetSampleCounter(pChannel);
-
-		wcmFilterCoord(common,pChannel,&ds);
-	}
-
-	/* skip event if we don't have enough movement */
-	suppress = wcmCheckSuppress(common, pLast, &ds);
-	if (suppress == SUPPRESS_ALL)
-		return;
-
 	/* JEJ - Do not move this code without discussing it with me.
 	 * The device state is invariant of any filtering performed below.
 	 * Changing the device state after this point can and will cause
@@ -1010,7 +1069,7 @@ void wcmEvent(WacomCommonPtr common, unsigned int channel,
 	/* For touch, only first finger moves the cursor */
 	if ((common->wcmTouch && ds.device_type == TOUCH_ID && ds.serial_num == 1) ||
 	    (ds.device_type != TOUCH_ID))
-		commonDispatchDevice(pInfo, pChannel, suppress);
+		commonDispatchDevice(pInfo, pChannel);
 }
 
 
@@ -1045,7 +1104,7 @@ rebasePressure(const WacomDevicePtr priv, const WacomDeviceState *ds)
 
 /**
  * Instead of reporting the raw pressure, we normalize
- * the pressure from 0 to FILTER_PRESSURE_RES. This is
+ * the pressure from 0 to maxCurve. This is
  * mainly to deal with the case where heavily used
  * stylus may have a "pre-loaded" initial pressure. To
  * do so, we keep the in-prox pressure and subtract it
@@ -1070,14 +1129,14 @@ normalizePressure(const WacomDevicePtr priv, const int raw_pressure)
 		p -= priv->minPressure;
 		range_left -= priv->minPressure;
 	}
-	/* normalize pressure to 0..FILTER_PRESSURE_RES */
+	/* normalize pressure to 0..maxCurve */
 	if (range_left >= 1)
 		pressure = xf86ScaleAxis(p,
-					 FILTER_PRESSURE_RES, 0,
+					 priv->maxCurve, 0,
 					 range_left,
 					 0);
 	else
-		pressure = FILTER_PRESSURE_RES;
+		pressure = priv->maxCurve;
 
 	return (int)pressure;
 }
@@ -1106,8 +1165,8 @@ setPressureButton(const WacomDevicePtr priv, int buttons, const int pressure)
 		{
 			/* don't set it off if it is within the tolerance
 			   and threshold is larger than the tolerance */
-			if ((common->wcmThreshold > THRESHOLD_TOLERANCE) &&
-			    (pressure > common->wcmThreshold - THRESHOLD_TOLERANCE))
+			if ((common->wcmThreshold > (priv->maxCurve * THRESHOLD_TOLERANCE)) &&
+			    (pressure > common->wcmThreshold - (priv->maxCurve * THRESHOLD_TOLERANCE)))
 				buttons |= button;
 		}
 	}
@@ -1157,13 +1216,13 @@ static void detectPressureIssue(WacomDevicePtr priv,
 }
 
 static void commonDispatchDevice(InputInfoPtr pInfo,
-				 const WacomChannelPtr pChannel,
-				 enum WacomSuppressMode suppress)
+				 const WacomChannelPtr pChannel)
 {
 	WacomDeviceState* ds = &pChannel->valid.states[0];
 	WacomDevicePtr priv = pInfo->private;
 	WacomCommonPtr common = priv->common;
 	WacomDeviceState filtered;
+	enum WacomSuppressMode suppress;
 	int raw_pressure = 0;
 
 	/* device_type should have been retrieved and set in the respective
@@ -1185,7 +1244,7 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 	if (priv->serial && filtered.serial_num != priv->serial)
 	{
 		DBG(10, priv, "serial number"
-			" is %u but your system configured %u",
+			" is %u but your system configured %u\n",
 			filtered.serial_num, priv->serial);
 		return;
 	}
@@ -1239,16 +1298,31 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 		}
 		filtered.pressure = applyPressureCurve(priv,&filtered);
 	}
-	else if (IsCursor(priv) && !priv->oldCursorHwProx)
+
+	/* Optionally filter values only while in proximity */
+	if (filtered.proximity && filtered.device_type != PAD_ID)
 	{
-		/* initial current max distance for Intuos series */
-		if ((TabletHasFeature(common, WCM_ROTATION)) ||
-				(TabletHasFeature(common, WCM_DUALINPUT)))
-			common->wcmMaxCursorDist = common->wcmMaxDist;
-		else
-			common->wcmMaxCursorDist = 0;
+		/* Start filter fresh when entering proximity */
+		if (!priv->oldState.proximity)
+			wcmResetSampleCounter(pChannel);
+
+		/* Reset filter whenever the tip is touched to the
+		 * screen to ensure clicks are sent from the pen's
+		 * actual position. Don't reset on other buttons or
+		 * tip-up, or else there may be a noticible jump/
+		 * hook produced in the middle/end of the stroke.
+		 */
+		if ((filtered.buttons & PRESSURE_BUTTON) && !(priv->oldState.buttons & PRESSURE_BUTTON))
+			wcmResetSampleCounter(pChannel);
+
+		wcmFilterCoord(common,pChannel,&filtered);
 	}
 
+	/* skip event if we don't have enough movement */
+	suppress = wcmCheckSuppress(common, &priv->oldState, &filtered);
+	if (suppress == SUPPRESS_ALL)
+		return;
+
 	/* Store cursor hardware prox for next use */
 	if (IsCursor(priv))
 		priv->oldCursorHwProx = ds->proximity;
@@ -1284,20 +1358,15 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 	/* force out-prox when distance is outside wcmCursorProxoutDist for pucks */
 	if (IsCursor(priv))
 	{
-		if (common->wcmProtocolLevel == WCM_PROTOCOL_5)
-		{
-			/* protocol 5 distance starts from the MaxDist
-			 * when getting in the prox.
-			 */
-			if (common->wcmMaxCursorDist > filtered.distance)
-				common->wcmMaxCursorDist = filtered.distance;
-		}
-		else
-		{
-			/* protocol 4 distance is 0 when getting in the prox */
-			if (common->wcmMaxCursorDist < filtered.distance)
-				common->wcmMaxCursorDist = filtered.distance;
-		}
+		/* Assume the the user clicks the puck buttons while
+		 * it is resting on the tablet. This works for both
+		 * tablets that have a normal distance scale (protocol
+		 * 5) as well as those with an inverted scale (protocol
+		 * 4 for many many kernel versions).
+		 */
+		if (filtered.buttons)
+			common->wcmMaxCursorDist = filtered.distance;
+
 		DBG(10, common, "Distance over"
 				" the tablet: %d, ProxoutDist: %d current"
 				" min/max %d hard prox: %d\n",
@@ -1306,21 +1375,23 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 				common->wcmMaxCursorDist,
 				ds->proximity);
 
-		if (priv->oldState.proximity)
-		{
-			if (abs(filtered.distance - common->wcmMaxCursorDist)
-					> common->wcmCursorProxoutDist)
-				filtered.proximity = 0;
-		}
-		/* once it is out. Don't let it in until a hard in */
-		/* or it gets inside wcmCursorProxoutDist */
-		else
-		{
-			if (abs(filtered.distance - common->wcmMaxCursorDist) >
-					common->wcmCursorProxoutDist && ds->proximity)
-				return;
-			if (!ds->proximity)
-				return;
+		if (common->wcmMaxCursorDist) {
+			if (priv->oldState.proximity)
+			{
+				if (abs(filtered.distance - common->wcmMaxCursorDist)
+						> common->wcmCursorProxoutDist)
+					filtered.proximity = 0;
+			}
+			/* once it is out. Don't let it in until a hard in */
+			/* or it gets inside wcmCursorProxoutDist */
+			else
+			{
+				if (abs(filtered.distance - common->wcmMaxCursorDist) >
+						common->wcmCursorProxoutDist && ds->proximity)
+					return;
+				if (!ds->proximity)
+					return;
+			}
 		}
 	}
 	wcmSendEvents(pInfo, &filtered);
@@ -1351,12 +1422,22 @@ int wcmInitTablet(InputInfoPtr pInfo, const char* id, float version)
 	if (common->wcmThreshold <= 0 && IsPen(priv))
 	{
 		/* Threshold for counting pressure as a button */
-		common->wcmThreshold = DEFAULT_THRESHOLD;
+		common->wcmThreshold = priv->maxCurve * DEFAULT_THRESHOLD;
 
 		xf86Msg(X_PROBED, "%s: using pressure threshold of %d for button 1\n",
 			pInfo->name, common->wcmThreshold);
 	}
 
+	/* Calculate default panscroll threshold if not set */
+	xf86Msg(X_CONFIG, "%s: panscroll is %d\n", pInfo->name, common->wcmPanscrollThreshold);
+	if (common->wcmPanscrollThreshold < 1) {
+		common->wcmPanscrollThreshold = common->wcmResolY * 13 / 1000; /* 13mm */
+	}
+	if (common->wcmPanscrollThreshold < 1) {
+		common->wcmPanscrollThreshold = 1000;
+	}
+	xf86Msg(X_CONFIG, "%s: panscroll modified to %d\n", pInfo->name, common->wcmPanscrollThreshold);
+
 	/* output tablet state as probed */
 	if (IsPen(priv))
 		xf86Msg(X_PROBED, "%s: maxX=%d maxY=%d maxZ=%d "
@@ -1402,10 +1483,13 @@ static int applyPressureCurve(WacomDevicePtr pDev, const WacomDeviceStatePtr pSt
 	/* clip the pressure */
 	int p = max(0, pState->pressure);
 
-	p = min(FILTER_PRESSURE_RES, p);
+	p = min(pDev->maxCurve, p);
 
 	/* apply pressure curve function */
-	return pDev->pPressCurve[p];
+	if (pDev->pPressCurve == NULL)
+		return p;
+	else
+		return pDev->pPressCurve[p];
 }
 
 /*****************************************************************************
@@ -1466,6 +1550,7 @@ WacomCommonPtr wcmNewCommon(void)
 			/* transmit position if increment is superior */
 	common->wcmRawSample = DEFAULT_SAMPLES;
 			/* number of raw data to be used to for filtering */
+	common->wcmPanscrollThreshold = 0;
 	common->wcmPressureRecalibration = 1;
 	return common;
 }
@@ -1475,6 +1560,9 @@ void wcmFreeCommon(WacomCommonPtr *ptr)
 {
 	WacomCommonPtr common = *ptr;
 
+	if (!common)
+		return;
+
 	DBG(10, common, "common refcount dec to %d\n", common->refcnt - 1);
 	if (--common->refcnt == 0)
 	{
