@@ -99,6 +99,7 @@ static Atom prop_tooltype;
 static Atom prop_btnactions;
 static Atom prop_product_id;
 static Atom prop_pressure_recal;
+static Atom prop_panscroll_threshold;
 #ifdef DEBUG
 static Atom prop_debuglevels;
 #endif
@@ -106,21 +107,23 @@ static Atom prop_debuglevels;
 /**
  * Calculate a user-visible pressure level from a driver-internal pressure
  * level. Pressure settings exposed to the user assume a range of 0-2047
- * while the driver scales everything to a range of 0-FILTER_PRESSURE_RES.
+ * while the driver scales everything to a range of 0-maxCurve.
  */
-static inline int wcmInternalToUserPressure(int pressure)
+static inline int wcmInternalToUserPressure(InputInfoPtr pInfo, int pressure)
 {
-	return pressure / (FILTER_PRESSURE_RES / 2048);
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	return pressure / (priv->maxCurve / 2048);
 }
 
 /**
  * Calculate a driver-internal pressure level from a user-visible pressure
  * level. Pressure settings exposed to the user assume a range of 0-2047
- * while the driver scales everything to a range of 0-FILTER_PRESSURE_RES.
+ * while the driver scales everything to a range of 0-maxCurve.
  */
-static inline int wcmUserToInternalPressure(int pressure)
+static inline int wcmUserToInternalPressure(InputInfoPtr pInfo, int pressure)
 {
-	return pressure * (FILTER_PRESSURE_RES / 2048);
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	return pressure * (priv->maxCurve / 2048);
 }
 
 /**
@@ -276,7 +279,7 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	}
 
 	values[0] = (!common->wcmMaxZ) ? 0 : common->wcmThreshold;
-	values[0] = wcmInternalToUserPressure(values[0]);
+	values[0] = wcmInternalToUserPressure(pInfo, values[0]);
 	prop_threshold = InitWcmAtom(pInfo->dev, WACOM_PROP_PRESSURE_THRESHOLD, XA_INTEGER, 32, 1, values);
 
 	values[0] = common->wcmSuppress;
@@ -334,6 +337,9 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 						  XA_INTEGER, 8, 1, values);
 	}
 
+	values[0] = common->wcmPanscrollThreshold;
+	prop_panscroll_threshold = InitWcmAtom(pInfo->dev, WACOM_PROP_PANSCROLL_THRESHOLD, XA_INTEGER, 32, 1, values);
+
 	values[0] = common->vendor_id;
 	values[1] = common->tablet_id;
 	prop_product_id = InitWcmAtom(pInfo->dev, XI_PROP_PRODUCT_ID, XA_INTEGER, 32, 2, values);
@@ -453,6 +459,8 @@ static int wcmCheckActionProperty(WacomDevicePtr priv, Atom property, XIProperty
 				break;
 			case AC_MODETOGGLE:
 				break;
+			case AC_PANSCROLL:
+				break;
 			default:
 				DBG(3, priv, "ERROR: Unknown command\n");
 				return BadValue;
@@ -846,7 +854,7 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			common->wcmCursorProxoutDist = value;
 	} else if (property == prop_threshold)
 	{
-		const INT32 MAXIMUM = wcmInternalToUserPressure(FILTER_PRESSURE_RES);
+		const INT32 MAXIMUM = wcmInternalToUserPressure(pInfo, priv->maxCurve);
 		INT32 value;
 
 		if (prop->size != 1 || prop->format != 32)
@@ -855,11 +863,11 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 		value = *(INT32*)prop->data;
 
 		if (value == -1)
-			value = DEFAULT_THRESHOLD;
+			value = priv->maxCurve * DEFAULT_THRESHOLD;
 		else if ((value < 1) || (value > MAXIMUM))
 			return BadValue;
 		else
-			value = wcmUserToInternalPressure(value);
+			value = wcmUserToInternalPressure(pInfo, value);
 
 		if (!checkonly)
 			common->wcmThreshold = value;
@@ -970,6 +978,21 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 
 		if (!checkonly)
 			common->wcmPressureRecalibration = values[0];
+	} else if (property == prop_panscroll_threshold)
+	{
+		CARD32 *values = (CARD32*)prop->data;
+
+		if (prop->size != 1 || prop->format != 32)
+			return BadValue;
+
+		if (values[0] <= 0)
+			return BadValue;
+
+		if (IsTouch(priv))
+			return BadMatch;
+
+		if (!checkonly)
+			common->wcmPanscrollThreshold = values[0];
 	} else
 	{
 		Atom *handler = NULL;
