@@ -471,6 +471,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_BOOLEAN
 	},
+	{
+		.name = "PanScrollThreshold",
+		.x11name = "PanScrollThreshold",
+		.desc = "Adjusts distance required for pan actions to generate a scroll event",
+		.prop_name = WACOM_PROP_PANSCROLL_THRESHOLD,
+		.prop_format = 32,
+		.prop_offset = 0,
+		.arg_count = 1,
+	},
 	{
 		.name = "MapToOutput",
 		.desc = "Map the device to the given output. ",
@@ -484,7 +493,7 @@ static param_t parameters[] =
 		.get_func = get_all,
 		.prop_flags = PROP_FLAG_READONLY,
 	},
-	{}
+	{.name = NULL}
 };
 
 /**
@@ -581,6 +590,8 @@ static struct modifier modifiers[] = {
 	{"lhyper", "Hyper_L"},
 	{"rhyper", "Hyper_R"},
 
+	{"altgr", "ISO_Level3_Shift"},
+
 	{ NULL, NULL }
 };
 
@@ -605,6 +616,8 @@ static struct modifier specialkeys[] = {
 
 	{"backspace", "BackSpace"}, {"Backspace", "BackSpace"},
 
+	{"return", "Return"},
+
 	{"tab", "Tab"},
 
 	{"PgUp", "Prior"}, {"PgDn", "Next"},
@@ -749,8 +762,7 @@ static void usage(void)
 
 static void version(void)
 {
-	printf("%d.%d.%d\n", PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR,
-			     PACKAGE_VERSION_PATCHLEVEL);
+	printf(BUILD_VERSION "\n");
 }
 
 static XDevice* find_device(Display *display, char *name)
@@ -937,6 +949,9 @@ static void list_mod(Display *dpy)
 	m = specialkeys;
 	while(m->name)
 		printf("	%s\n", m++->name);
+
+	printf("Keys not listed above can be specified via their KeySyms."
+		" See the man page for details.\n");
 }
 
 static void list(Display *dpy, int argc, char **argv)
@@ -978,7 +993,7 @@ static const char *convert_specialkey(const char *specialkey)
 			m++;
 	}
 
-	return m->converted ? m->converted : (char*)specialkey;
+	return m->converted ? m->converted : specialkey;
 }
 
 /**
@@ -1005,6 +1020,7 @@ static int special_map_button(Display *dpy, int argc, char **argv, unsigned long
 static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
+static int special_map_panscroll(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 
 /* Valid keywords for the --set ButtonX options */
 static struct keywords {
@@ -1016,6 +1032,7 @@ static struct keywords {
 	{"core", special_map_core},
 	{"modetoggle", special_map_modetoggle},
 	{"displaytoggle", special_map_displaytoggle},
+	{"pan", special_map_panscroll},
 	{ NULL, NULL }
 };
 
@@ -1026,7 +1043,7 @@ static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *
 	static int once_only = 1;
 	if (once_only)
 	{
-		printf ("Note: The \"core\" keyword is not supported anymore and "
+		fprintf(stderr, "Note: The \"core\" keyword is not supported anymore and "
 			"will be ignored.\n");
 		once_only = 0;
 	}
@@ -1053,13 +1070,26 @@ static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsign
 	static int once_only = 1;
 	if (once_only)
 	{
-		printf ("Note: The \"displaytoggle\" keyword is not supported "
+		fprintf(stderr, "Note: The \"displaytoggle\" keyword is not supported "
 			"anymore and will be ignored.\n");
 		once_only = 0;
 	}
 	return 0;
 }
 
+static int special_map_panscroll(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
+{
+	if (*ndata + 1 > size) {
+		fprintf(stderr, "Insufficient space to store all commands.\n");
+		return 0;
+	}
+	data[*ndata] = AC_PANSCROLL;
+
+	*ndata += 1;
+
+	return 0;
+}
+
 static inline int is_valid_keyword(const char *keyword)
 {
 	struct keywords *kw = keywords;
@@ -1168,9 +1198,12 @@ static int keysym_to_keycode(Display *dpy, KeySym sym)
 	XkbStateRec state;
 	int kc = 0;
 
-
 	if (!xkb)
-		xkb = XkbGetKeyboard(dpy, XkbAllComponentsMask, XkbUseCoreKbd);
+		xkb = XkbGetMap(dpy, XkbAllComponentsMask, XkbUseCoreKbd);
+	if (!xkb) {
+		fprintf(stderr, "Warning: failed to query keyboard map\n");
+		goto out;
+	}
 	XkbGetState(dpy, XkbUseCoreKbd, &state);
 
 	for (kc = xkb->min_key_code; kc <= xkb->max_key_code; kc++)
@@ -1332,17 +1365,20 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 	int  i = 0;
 	int  nwords = 0;
 	char **words = NULL;
+	int n;
 	
 	/* translate cmdline commands */
 	words = strjoinsplit(argc, argv, &nwords);
 
-	if (nwords==1 && sscanf(words[0], "%d", &i) == 1)
+	if (nwords==1 && sscanf(words[0], "%d", &n) == 1)
 	{ /* Mangle "simple" button maps into proper actions */
 		char *nargv[1];
 		
+		for (i =  0; i < nwords; i++)
+			free(words[i]);
 		free(words);
 		nargv[0] = alloca(32);
-		sprintf(nargv[0], "button +%d", i);
+		sprintf(nargv[0], "button +%d", n);
 		words = strjoinsplit(1, nargv, &nwords);
 	}
 
@@ -1376,6 +1412,8 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 		}
 	}
 
+	for (i = 0; i < nwords; i++)
+		free(words[i]);
 	free(words);
 
 	return True;
@@ -1420,14 +1458,14 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 	if (offset >= btnact_nitems)
 	{
 		fprintf(stderr, "Invalid offset into %s property.\n", XGetAtomName(dpy, btnact_prop));
-		goto out;
+		goto out2;
 	}
 
 	if (format != 32 || type != XA_ATOM)
 	{
 		fprintf(stderr, "Property '%s' in an unexpected format. This is a bug.\n",
 		        XGetAtomName(dpy, btnact_prop));
-		goto out;
+		goto out2;
 	}
 
 	/* set or unset the property */
@@ -1441,7 +1479,7 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 		 */
 		fprintf(stderr, "Unsupported offset into '%s' property.\n",
 			XGetAtomName(dpy, btnact_prop));
-		goto out;
+		goto out2;
 	}
 
 	if (nitems > 0)
@@ -1468,6 +1506,8 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 	}
 
 	XFlush(dpy);
+out2:
+	XFree(btnact_data);
 out:
 	free(data);
 }
@@ -1510,7 +1550,7 @@ static void map_actions(Display *dpy, XDevice *dev, param_t* param, int argc, ch
 
 	if (strcmp(param->prop_name, WACOM_PROP_BUTTON_ACTIONS) == 0)
 	{
-		if (sscanf(argv[0], "%d", &offset) != 1)
+		if (sscanf(argv[0], "%d", &offset) != 1 || offset <= 0)
 		{
 			fprintf(stderr, "'%s' is not a valid button number.\n", argv[0]);
 			return;
@@ -1544,7 +1584,7 @@ static void set_xydefault(Display *dpy, XDevice *dev, param_t* param, int argc,
 	{
 		fprintf(stderr, "Property for '%s' not available.\n",
 			param->name);
-		goto out;
+		return;
 	}
 
 	XGetDeviceProperty(dpy, dev, prop, 0, 1000, False, AnyPropertyType,
@@ -1566,7 +1606,7 @@ static void set_xydefault(Display *dpy, XDevice *dev, param_t* param, int argc,
 				PropModeReplace, data, nitems);
 	XFlush(dpy);
 out:
-	free(data);
+	XFree(data);
 }
 
 static void set_mode(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
@@ -1644,14 +1684,15 @@ static void set_rotate(Display *dpy, XDevice *dev, param_t* param, int argc, cha
 	{
 		fprintf(stderr, "Property for '%s' has no or wrong value - this is a bug.\n",
 			param->name);
-		return;
+		goto out;
 	}
 
 	*data = rotation;
 	XChangeDeviceProperty(dpy, dev, prop, type, format,
 				PropModeReplace, data, nitems);
 	XFlush(dpy);
-
+out:
+	XFree(data);
 	return;
 }
 
@@ -1827,7 +1868,7 @@ out:
 		free(values[i]);
 	free(values);
 	XCloseDevice(dpy, dev);
-	free(data);
+	XFree(data);
 }
 
 static void get_mode(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
@@ -1898,7 +1939,7 @@ static void get_rotate(Display *dpy, XDevice *dev, param_t* param, int argc, cha
 	{
 		fprintf(stderr, "Property for '%s' has no or wrong value - this is a bug.\n",
 			param->name);
-		return;
+		goto out;
 	}
 
 	switch(*data)
@@ -1919,6 +1960,8 @@ static void get_rotate(Display *dpy, XDevice *dev, param_t* param, int argc, cha
 
 	print_value(param, "%s", rotation);
 
+out:
+	XFree(data);
 	return;
 }
 
@@ -1941,6 +1984,7 @@ static int get_actions(Display *dpy, XDevice *dev,
 	unsigned long nitems, bytes_after, *data;
 	int i;
 	char buff[1024] = {0};
+	int last_type;
 
 	prop = XInternAtom(dpy, param->prop_name, True);
 
@@ -1969,9 +2013,9 @@ static int get_actions(Display *dpy, XDevice *dev,
 		   AnyPropertyType, &type, &format, &nitems,
 		   &bytes_after, (unsigned char**)&data);
 
+	last_type = 0;
 	for (i = 0; i < nitems; i++)
 	{
-		static int last_type;
 		unsigned long action = data[i];
 		int current_type;
 		int detail;
@@ -2011,7 +2055,7 @@ static int get_actions(Display *dpy, XDevice *dev,
 		if (current_type == AC_KEY)
 			sprintf(str, "%c%s ", press_str,
 				XKeysymToString(detail));
-		else
+		else if (current_type == AC_BUTTON)
 			sprintf(str, "%c%d ", press_str, detail);
 		strcat(buff, str);
 		last_type = current_type;
@@ -2101,7 +2145,7 @@ static void get_map(Display *dpy, XDevice *dev, param_t *param, int argc, char**
 
 	if (strcmp(param->prop_name, WACOM_PROP_BUTTON_ACTIONS) == 0)
 	{
-		if (sscanf(argv[0], "%d", &offset) != 1)
+		if (sscanf(argv[0], "%d", &offset) != 1 || offset <= 0)
 		{
 			fprintf(stderr, "'%s' is not a valid button number.\n", argv[0]);
 			return;
@@ -2185,9 +2229,10 @@ static Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height,
 {
 	Atom matrix_prop = XInternAtom(dpy, "Coordinate Transformation Matrix", True);
 	Atom type;
-	int format;
+	int format, i;
 	unsigned long nitems, bytes_after;
-	float *data;
+	unsigned long *data;
+	float matrix[9];
 	Bool matrix_is_valid = True;
 
 	int display_width = DisplayWidth(dpy, DefaultScreen(dpy));
@@ -2212,17 +2257,22 @@ static Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height,
 		return False;
 	}
 
-	TRACE("Current transformation matrix:\n");
-	TRACE("	[ %f %f %f ]\n", data[0], data[1], data[2]);
-	TRACE("	[ %f %f %f ]\n", data[3], data[4], data[5]);
-	TRACE("	[ %f %f %f ]\n", data[6], data[7], data[8]);
+	/* XI1 stores 32 bit properties (including float) as long,
+	 * regardless of architecture */
+	for (i = 0; i < ARRAY_SIZE(matrix); i++)
+		matrix[i] = *(float*)(&data[i]);
 
-	*width  = rint(display_width  * data[0]);
-	*x_org  = rint(display_width  * data[2]);
-	*height = rint(display_height * data[4]);
-	*y_org  = rint(display_height * data[5]);
-	if ((data[1] != 0 || data[3] != 0 || data[6] != 0 || data[7] != 0) ||
-	    (data[8] != 1)) {
+	TRACE("Current transformation matrix:\n");
+	TRACE("	[ %f %f %f ]\n", matrix[0], matrix[1], matrix[2]);
+	TRACE("	[ %f %f %f ]\n", matrix[3], matrix[4], matrix[5]);
+	TRACE("	[ %f %f %f ]\n", matrix[6], matrix[7], matrix[8]);
+
+	*width  = rint(display_width  * matrix[0]);
+	*x_org  = rint(display_width  * matrix[2]);
+	*height = rint(display_height * matrix[4]);
+	*y_org  = rint(display_height * matrix[5]);
+	if ((matrix[1] != 0 || matrix[3] != 0 || matrix[6] != 0 || matrix[7] != 0) ||
+	    (matrix[8] != 1)) {
 		fprintf(stderr, "Non-rectangular transformation matrix detected.\n");
 		matrix_is_valid = False;
 	}
@@ -2515,9 +2565,8 @@ out:
 static void set_output(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv)
 {
 	int head_no;
-	int x, y;
+	int x, y, flags;
 	unsigned int width, height;
-	int flags = XParseGeometry(argv[0], &x, &y, &width, &height);
 
 	if (argc != param->arg_count)
 	{
@@ -2526,6 +2575,8 @@ static void set_output(Display *dpy, XDevice *dev, param_t *param, int argc, cha
 		return;
 	}
 
+	flags = XParseGeometry(argv[0], &x, &y, &width, &height);
+
 	if (MaskIsSet(flags, XValue|YValue|WidthValue|HeightValue))
 		set_output_area(dpy, dev, x, y, width, height);
 	else if (strcasecmp(argv[0], "next") == 0)
@@ -2585,14 +2636,14 @@ static void get(Display *dpy, enum printformat printformat, int argc, char **arg
 	param_t *param;
 	XDevice *dev = NULL;
 
-	TRACE("'get' requested for '%s'.\n", argv[0]);
-
 	if (argc < 2)
 	{
 		usage();
 		return;
 	}
 
+	TRACE("'get' requested for '%s'.\n", argv[0]);
+
 	dev = find_device(dpy, argv[0]);
 	if (!dev)
 	{
@@ -2606,11 +2657,11 @@ static void get(Display *dpy, enum printformat printformat, int argc, char **arg
 		if (is_deprecated_parameter(argv[1]))
 			return;
 		fprintf(stderr, "Unknown parameter name '%s'.\n", argv[1]);
-		return;
+		goto out;
 	} else if (param->prop_flags & PROP_FLAG_WRITEONLY)
 	{
 		fprintf(stderr, "'%s' is a write-only option.\n", argv[1]);
-		return;
+		goto out;
 	} else
 	{
 		param->printformat = printformat;
@@ -2619,6 +2670,7 @@ static void get(Display *dpy, enum printformat printformat, int argc, char **arg
 
 	get_param(dpy, dev, param, argc - 2, &argv[2]);
 
+out:
 	XCloseDevice(dpy, dev);
 }
 
@@ -2657,7 +2709,7 @@ static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char
 	if (nitems <= param->prop_offset)
 	{
 		fprintf(stderr, "Property offset doesn't exist.\n");
-		return;
+		goto out;
 	}
 
 
@@ -2693,10 +2745,56 @@ static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char
 			print_value(param, "%s", str);
 			break;
 	}
+
+out:
+	XFree(data);
 }
 
 
 #ifndef BUILD_TEST
+
+#ifdef BUILD_FUZZINTERFACE
+void argsfromstdin(int *argc, char ***argv)
+{
+	const int READSIZE = 256;
+	char *buf = strdup((*argv)[0]);
+	size_t len = strlen(buf)+1;
+
+	while (1) {
+		char *p = realloc(buf, len + READSIZE);
+		size_t n;
+
+		if (!p)
+			exit(1);
+		buf = p;
+
+		n = fread(buf+len, 1, READSIZE, stdin);
+		if (n > 0) {
+			len += n;
+		}
+		else {
+			buf[len] = '\0';
+			len++;
+			break;
+		}
+	}
+
+	*argc = 0;
+	*argv = NULL;
+	while (len) {
+		char **p = realloc(*argv, (*argc + 1) * sizeof(char**));
+		if (!p)
+			exit(1);
+		*argv = p;
+		(*argv)[*argc] = buf;
+		(*argc)++;
+
+		len -= strlen(buf) + 1;
+		buf += strlen(buf) + 1;
+	}
+}
+#endif /* BUILD_FUZZINTERFACE */
+
 int main (int argc, char **argv)
 {
 	int c;
@@ -2719,6 +2817,10 @@ int main (int argc, char **argv)
 		{NULL, 0, NULL, 0}
 	};
 
+#ifdef BUILD_FUZZINTERFACE
+	argsfromstdin(&argc, &argv);
+#endif /* BUILD FUZZINTERFACE */
+
 	if (argc < 2)
 	{
 		usage();
@@ -2900,7 +3002,7 @@ static void test_parameter_number(void)
 	 * deprecated them.
 	 * Numbers include trailing NULL entry.
 	 */
-	assert(ARRAY_SIZE(parameters) == 39);
+	assert(ARRAY_SIZE(parameters) == 40);
 	assert(ARRAY_SIZE(deprecated_parameters) == 17);
 }
 
